# Title: Constrained Optimization Approaches for Estimation of Structural Models
#
# Authors: Che-Lin Su and Kenneth L. Judd, 2008
# 
#  
#####  HAROLD ZURCHER BUS REPAIR EXAMPLE  #####
#
# A constrained optimization formulation to compute maximum likelihood estimates 
# of the Harold Zurcher bus problem in Rust, Econometrica, 1987. 
#
#####  SET UP THE PROBLEM  #####
#
#  Define and process the data
param nBusG1;             # number of buses in Group 1 in the data
set B1 := 1..nBusG1;     # B1 is the index set of buses in Group 1
param nTG1;		# number of periods in Group 1 in the data
set T1 := 1..nTG1;	# T1 is the vector of time indices in Group 1

param nBusG2;             # number of buses in Group 1 in the data
set B2 := 1..nBusG2;     # B1 is the index set of buses in Group 1
param nTG2;		# number of periods in Group 1 in the data
set T2 := 1..nTG2;	# T1 is the vector of time indices in Group 1

param nBusG3;             # number of buses in Group 1 in the data
set B3 := 1..nBusG3;     # B1 is the index set of buses in Group 1
param nTG3;		# number of periods in Group 1 in the data
set T3 := 1..nTG3;	# T1 is the vector of time indices in Group 1

param nBusG4;             # number of buses in Group 1 in the data
set B4 := 1..nBusG4;     # B1 is the index set of buses in Group 1
param nTG4;		# number of periods in Group 1 in the data
set T4 := 1..nTG4;	# T1 is the vector of time indices in Group 1

param G1_XtReal {T1, B1}; 	# G1_XtReal[t, b] is the true mileage of bus b at time t in Group 1
param G2_XtReal {T2, B2}; 	# G2_XtReal[t, b] is the true mileage of bus b at time t in Group 2
param G3_XtReal {T3, B3}; 	# G3_XtReal[t, b] is the true mileage of bus b at time t in Group 3
param G4_XtReal {T4, B4}; 	# G4_XtReal[t, b] is the true mileage of bus b at time t in Group 4

param xscale; # scaling factor on the real mileage data

param G1_Xt {t in T1, b in B1} := G1_XtReal[t,b]/xscale; # G1_Xt[t, b] is the scaled mileage of bus b at time t in Group 1
param G2_Xt {t in T2, b in B2} := G2_XtReal[t,b]/xscale; # G2_Xt[t, b] is the scaled mileage of bus b at time t in Group 2
param G3_Xt {t in T3, b in B3} := G3_XtReal[t,b]/xscale; # G3_Xt[t, b] is the scaled mileage of bus b at time t in Group 3
param G4_Xt {t in T4, b in B4} := G4_XtReal[t,b]/xscale; # G4_Xt[t, b] is the scaled mileage of bus b at time t in Group 4

#  Define the state space used in the dynamic programming part
param    N; 	# number of states used in dynamic programming approximation
set X := 1..N; 	# X is the index set of states

param xmin; # minimum mileage reading = xmin * xscale;
param xmax; # maximum mileage reading = xmax * xscale;
param x {i in X} := xmin + (xmax-xmin)/(N-1)*(i-1); 
#  x[i] denotes state i; the set of states is a uniform grid on the interval [xmin, xmax]

param G1_dt {T1, B1};	# decision of bus b at time t in Group 1
param G2_dt {T2, B2};	# decision of bus b at time t in Group 2
param G3_dt {T3, B3};	# decision of bus b at time t in Group 3
param G4_dt {T4, B4};	# decision of bus b at time t in Group 4

#  The dynamic programming model in the estimation lives on a discrete state
#  Binning process: assign scaled mileage Xt[t] to the closest state in X
param G1_xt {t in T1, b in B1} := ceil(G1_Xt[t, b]/(xmax-xmin)*(N-1)+0.5); 
param G2_xt {t in T2, b in B2} := ceil(G2_Xt[t, b]/(xmax-xmin)*(N-1)+0.5); 
param G3_xt {t in T3, b in B3} := ceil(G3_Xt[t, b]/(xmax-xmin)*(N-1)+0.5); 
param G4_xt {t in T4, b in B4} := ceil(G4_Xt[t, b]/(xmax-xmin)*(N-1)+0.5); 
#
#  Define "known" structural parameters
#  We fix beta since data cannot identify it
param beta;      	# discount factor
#
#####  DECLARE STRUCTURAL PARAMETERS TO BE ESTIMATED #####
#####  Parameters for cost function  #####
#  c(x, thetaCost) = thetaCost[1]*x + thetaCost[2]*x^0.5
var thetaCost {1..2} >= 0; 
#
#####  Parameters and definition of transition process  #####
#  thetaProbs defines Markov chain
var thetaProbs {1..3} >= 0; 
#
#  Define the Markov chain representing the changes in mileage on the x[i] grid.
#
#  The state increases by some amount in [0,JumpMax] where 
#  JumpMax is the maximum increase in mileage in one period
#  and equals a fraction JumpRatio of the range of mileage in the state space
param JumpRatio;             
param JumpMax := (xmax-xmin) * JumpRatio;
#
#  We assume that the jumps are independent of the current state
#  and its transition process has three pieces, each a uniform distribution    
#  Define 1st break point for stepwise uniform distribution in mileage increase
param M1 := ceil(1/4*JumpMax/(xmax-xmin)*(N-1)+0.5);   
#  Define 2nd break point for stepwise uniform distribution in mileage increase
param M2 := ceil(3/4*JumpMax/(xmax-xmin)*(N-1)+0.5);   
#  Define end point for stepwise uniform distribution in mileage increase
param M := ceil(JumpMax/(xmax-xmin)*(N-1)+0.5);     
 
#
#  Y is the vector of elements in transition rule
set Y := 1..M;
var TransProb {i in Y} = 
	if 	i <= M1 		then 	thetaProbs[1]/M1 
	else if 	i > M1 and i <= M2 	then 	thetaProbs[2]/(M2-M1) 
	else 	thetaProbs[3]/(M-M2);
#
#####  Scrap value parameter  #####
var RC >= 0;
##### END OF STRUCTURAL VARIABLES  #####
#
#####  DECLARE EQUILIBRIUM CONSTRAINT VARIABLES #####
# The NLP approach requires us to solve equilibrium constraint variables
var EV {X};        		# Value Function of each state
#####  END OF EQUILIBRIUM CONSTRAINT VARIABLES  #####
#
#####  DECLARE AUXILIARY VARIABLES  ######
#  Define auxiliary variables to economize on expressions	
#  Create Cost variable to represent the cost function; 
#  Cost[i] is the cost of regular maintenance at x[i].
var Cost {i in X} = sum {j in 1..2} thetaCost[j]*x[i]^(1/j);	    
#  Let CbEV[i] represent - Cost[i] + beta*EV[i]; 
#  this is the expected payoff at x[i] if regular maintenance is chosen
var CbEV {i in X} = - Cost[i] + beta*EV[i];                     
#  Let PayoffDiff[i] represent -CbEV[i] - RC + CbEV[1]; 
#  this is the difference in expected payoff at x[i] between engine replacement and regular maintenance
var PayoffDiff {i in X} = -CbEV[i] - RC + CbEV[1];               
#  Let ProbRegMaint[i] represent 1/(1+exp(PayoffDiff[i])); 
#  this is the probability of performing regular maintenance at state x[i];
var ProbRegMaint {i in X} = 1/(1+exp(PayoffDiff[i]));     

var BellmanViola {i in 1..(N-M+1)} = sum {j in 0..(M-1)} log(exp(CbEV[i+j])+ exp(-RC + CbEV[1]))* TransProb[j+1] - EV[i];


#####  END OF AUXILIARY VARIABLES  #####
#
#####  OBJECTIVE AND CONSTRAINT DEFINITIONS  #####
#
## Define objective function
##
## # Likelihood function for Group 4
maximize Likelihood_G4: 
## #  The likelihood function contains two pieces
## #  First is the likelihood that the engine is replaced given time t state in the data.
	sum {t in 2..nTG4, b in B4} log(G4_dt[t,b]*(1-ProbRegMaint[G4_xt[t,b]]) + (1-G4_dt[t,b])*ProbRegMaint[G4_xt[t,b]])
## #  Second is the likelihood that the observed transition between t-1 and t would have occurred.
	+ sum {t in 2..nTG4, b in B4} log(G4_dt[t-1,b]*(TransProb[G4_xt[t,b]-1+1]) + (1-G4_dt[t-1,b])*(TransProb[G4_xt[t,b]-G4_xt[t-1,b]+1])); 

## # Likelihood function for Group 1 to Group 3
maximize Likelihood_G1to3: 
## #  The likelihood function contains two pieces for each group
## #  Group 1: First is the likelihood that the engine is replaced given time t state in the data.
	sum {t in 2..nTG1, b in B1} log(G1_dt[t,b]*(1-ProbRegMaint[G1_xt[t,b]]) + (1-G1_dt[t,b])*ProbRegMaint[G1_xt[t,b]])
## #  Group 1: Second is the likelihood that the observed transition between t-1 and t would have occurred.
	+ sum {t in 2..nTG1, b in B1} log(G1_dt[t-1,b]*(TransProb[G1_xt[t,b]-1+1]) + (1-G1_dt[t-1,b])*(TransProb[G1_xt[t,b]-G1_xt[t-1,b]+1])) 
## #  Group 2: First is the likelihood that the engine is replaced given time t state in the data.
	+ sum {t in 2..nTG2, b in B2} log(G2_dt[t,b]*(1-ProbRegMaint[G2_xt[t,b]]) + (1-G2_dt[t,b])*ProbRegMaint[G2_xt[t,b]])
## #  Group 2: Second is the likelihood that the observed transition between t-1 and t would have occurred.
	+ sum {t in 2..nTG2, b in B2} log(G2_dt[t-1,b]*(TransProb[G2_xt[t,b]-1+1]) + (1-G2_dt[t-1,b])*(TransProb[G2_xt[t,b]-G2_xt[t-1,b]+1])) 
## #  Group 3: First is the likelihood that the engine is replaced given time t state in the data.
	+ sum {t in 2..nTG3, b in B3} log(G3_dt[t,b]*(1-ProbRegMaint[G3_xt[t,b]]) + (1-G3_dt[t,b])*ProbRegMaint[G3_xt[t,b]])
## #  Group 3: Second is the likelihood that the observed transition between t-1 and t would have occurred.
	+ sum {t in 2..nTG3, b in B3} log(G3_dt[t-1,b]*(TransProb[G3_xt[t,b]-1+1]) + (1-G3_dt[t-1,b])*(TransProb[G3_xt[t,b]-G3_xt[t-1,b]+1])); 

## # Likelihood function for Group 1 to Group 4
maximize Likelihood_G1to4: 
## #  The likelihood function contains two pieces for each group
## #  Group 1: First is the likelihood that the engine is replaced given time t state in the data.
	sum {t in 2..nTG1, b in B1} log(G1_dt[t,b]*(1-ProbRegMaint[G1_xt[t,b]]) + (1-G1_dt[t,b])*ProbRegMaint[G1_xt[t,b]])
## #  Group 1: Second is the likelihood that the observed transition between t-1 and t would have occurred.
	+ sum {t in 2..nTG1, b in B1} log(G1_dt[t-1,b]*(TransProb[G1_xt[t,b]-1+1]) + (1-G1_dt[t-1,b])*(TransProb[G1_xt[t,b]-G1_xt[t-1,b]+1])) 
## #  Group 2: First is the likelihood that the engine is replaced given time t state in the data.
	+ sum {t in 2..nTG2, b in B2} log(G2_dt[t,b]*(1-ProbRegMaint[G2_xt[t,b]]) + (1-G2_dt[t,b])*ProbRegMaint[G2_xt[t,b]])
## #  Group 2: Second is the likelihood that the observed transition between t-1 and t would have occurred.
	+ sum {t in 2..nTG2, b in B2} log(G2_dt[t-1,b]*(TransProb[G2_xt[t,b]-1+1]) + (1-G2_dt[t-1,b])*(TransProb[G2_xt[t,b]-G2_xt[t-1,b]+1])) 
## #  Group 3: First is the likelihood that the engine is replaced given time t state in the data.
	+ sum {t in 2..nTG3, b in B3} log(G3_dt[t,b]*(1-ProbRegMaint[G3_xt[t,b]]) + (1-G3_dt[t,b])*ProbRegMaint[G3_xt[t,b]])
## #  Group 3: Second is the likelihood that the observed transition between t-1 and t would have occurred.
	+ sum {t in 2..nTG3, b in B3} log(G3_dt[t-1,b]*(TransProb[G3_xt[t,b]-1+1]) + (1-G3_dt[t-1,b])*(TransProb[G3_xt[t,b]-G3_xt[t-1,b]+1])) 
## #  Group 4: First is the likelihood that the engine is replaced given time t state in the data.
	+ sum {t in 2..nTG4, b in B4} log(G4_dt[t,b]*(1-ProbRegMaint[G4_xt[t,b]]) + (1-G4_dt[t,b])*ProbRegMaint[G4_xt[t,b]])
## #  Group 4: Second is the likelihood that the observed transition between t-1 and t would have occurred.
	+ sum {t in 2..nTG4, b in B4} log(G4_dt[t-1,b]*(TransProb[G4_xt[t,b]-1+1]) + (1-G4_dt[t-1,b])*(TransProb[G4_xt[t,b]-G4_xt[t-1,b]+1])); 

#
#  List the constraints
#
subject to 
#
#  Bellman equation for states below N-M
	Bellman_1toNminusM {i in X: i <= N-(M-1)}: 
		EV[i] = sum {j in 0..(M-1)} 
			log(exp(CbEV[i+j])+ exp(-RC + CbEV[1]))* TransProb[j+1];
#
#  Bellman equation for states above N-M, (we adjust transition probabilities to keep state in [xmin, xmax])
	Bellman_LastM {i in X: i > N-(M-1) and i <= N-1}: 
		EV[i] = (sum {j in 0..(N-i-1)} 
			log(exp(CbEV[i+j])+ exp(-RC + CbEV[1]))* TransProb[j+1]) 
			+ (1- sum {k in 0..(N-i-1)} TransProb[k+1]) * log(exp(CbEV[N])+ exp(-RC + CbEV[1]));
#
#  Bellman equation for state N
	Bellman_N:  EV[N] = log(exp(CbEV[N])+ exp(-RC + CbEV[1]));
#
#  The probability parameters in transition process must add to one
    Probability: sum {i in 1..3} thetaProbs[i] = 1;

#
#  Put bound on EV; this should not bind, but is a cautionary step to help keep algorithm within bounds
    EVBound {i in X}: EV[i] <= 500;
#

####   DEFINE THE PROBLEM   #####
#
# Name the problem
problem MPECZurcher_G4:
#
# Choose the objective function
Likelihood_G4,

#
# List the variables
EV, RC, thetaCost, thetaProbs, TransProb, Cost, CbEV, PayoffDiff, ProbRegMaint, BellmanViola, 

#
# List the constraints
Bellman_1toNminusM,
Bellman_LastM, 
Bellman_N, 
Probability,  
EVBound;

# Name the problem
problem MPECZurcher_G1to3:
#
# Choose the objective function
Likelihood_G1to3,

#
# List the variables
EV, RC, thetaCost, thetaProbs, TransProb, Cost, CbEV, PayoffDiff, ProbRegMaint, BellmanViola, 

#
# List the constraints
Bellman_1toNminusM,
Bellman_LastM, 
Bellman_N, 
Probability,  
EVBound;

# Name the problem
problem MPECZurcher_G1to4:
#
# Choose the objective function
Likelihood_G1to4,

#
# List the variables
EV, RC, thetaCost, thetaProbs, TransProb, Cost, CbEV, PayoffDiff, ProbRegMaint, BellmanViola, 

#
# List the constraints
Bellman_1toNminusM,
Bellman_LastM, 
Bellman_N, 
Probability,  
EVBound;


#
#################################
